@@ -2,7 +2,7 @@
  * SPDX-License-Identifier: GPL-2.0-or-later
  *
  * bittwist - pcap based ethernet packet generator
- * Copyright (C) 2006 - 2023 Addy Yeow Chin Heng <ayeowch@gmail.com>
+ * Copyright (C) 2006 - 2023 Addy Yeow <ayeowch@gmail.com>
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -29,29 +29,29 @@ int32_t thiszone; /* offset from GMT to local time in seconds */
 char ebuf[PCAP_ERRBUF_SIZE]; /* pcap error buffer */
 
 /* options */
-int vflag = 0;              /* 1 - print timestamp, 2 - print timestamp and hex data */
-int len = 0;                /* packet length to send (-1 = captured, 0 = on wire, or positive
-                               value <= 1514) */
-int interval = 0;           /* a constant interval in seconds (0 means actual interval
-                               will be used instead) */
-int linerate = -1;          /* limit packet throughput at the specified Mbps (0 means no limit) */
-uint64_t bps = 0;           /* bits per second converted from linerate */
-unsigned long max_pkts = 0; /* send up to the specified number of packets */
+int vflag = 0;           /* 1 = print timestamp, 2 = print timestamp and hex data */
+int len = 0;             /* packet length to send (-1 = captured, 0 = on wire, or positive
+                            value <= 1514) */
+int pps = 0;             /* packets per second (1 to 1000000) */
+int gap = 0;             /* gap/interval between packets in seconds (1 to 86400) */
+int linerate = -1;       /* limit packet throughput at the specified Mbps (0 means no limit) */
+uint64_t bps = 0;        /* bits per second converted from linerate */
+bool default_gap = true; /* true = use captured interval, false = custom interval */
+uint64_t max_pkts = 0;   /* send up to the specified number of packets */
 
 /* data */
-pcap_t *pd = NULL;         /* pcap descriptor */
-uint8_t *pkt_data = NULL;  /* packet data including the link-layer header */
-FILE **trace_files = NULL; /* pointers to trace files */
 int trace_files_count = 0;
-struct timespec sleep_ts = {0, 0}; /* sleep duration to shape throughput */
-struct timespec curr_ts;           /* timestamp of current packet */
-struct timespec prev_ts = {0, 0};  /* timestamp of previous packet */
-struct token_bucket tb = {0, 0};   /* token bucket for shaping throughput at linerate */
+trace_file_t **trace_files = NULL; /* pointers to trace files */
+pcap_t *pd = NULL;                 /* pcap descriptor */
+uint8_t pkt_data[ETH_MAX_LEN];     /* packet data including the link-layer header */
+struct pcap_sf_pkthdr header;      /* pcap header per packet */
+struct token_bucket tb_pps;        /* token bucket for shaping throughput at pps */
+struct token_bucket tb_linerate;   /* token bucket for shaping throughput at linerate */
 
 /* stats */
-static unsigned long pkts_sent = 0;
-static unsigned long bytes_sent = 0;
-static unsigned long failed = 0;
+static uint64_t pkts_sent = 0;
+static uint64_t bytes_sent = 0;
+static uint64_t failed = 0;
 struct timeval start = {0, 0};
 struct timeval end = {0, 0};
 
@@ -72,7 +72,7 @@ int main(int argc, char **argv)
         program_name = argv[0];
 
     /* process options */
-    while ((c = getopt(argc, argv, "dvi:s:l:c:p:r:h")) != -1)
+    while ((c = getopt(argc, argv, "dvi:s:l:c:p:t:r:h")) != -1)
     {
         switch (c)
         {
@@ -121,9 +121,8 @@ int main(int argc, char **argv)
             len = strtol(optarg, NULL, 0);
             if (len != -1 && len != 0)
             {
-                if (len < ETHER_HDR_LEN || len > ETHER_MAX_LEN)
-                    error("value for length must be between %d to %d", ETHER_HDR_LEN,
-                          ETHER_MAX_LEN);
+                if (len < ETH_HDR_LEN || len > ETH_MAX_LEN)
+                    error("value for length must be between %d to %d", ETH_HDR_LEN, ETH_MAX_LEN);
             }
             break;
         case 'l':
@@ -133,16 +132,21 @@ int main(int argc, char **argv)
             max_pkts = strtoul(optarg, NULL, 0); /* send all packets if max_pkts <= 0 */
             break;
         case 'p':
-            interval = strtol(optarg, NULL, 0);
-            if (interval < 1 || interval > SLEEP_MAX)
-                error("value for sleep must be between 1 to %d", SLEEP_MAX);
+            pps = strtol(optarg, NULL, 0);
+            if (pps < 1 || pps > PPS_MAX)
+                error("value for pps must be between 1 to %d", PPS_MAX);
+            break;
+        case 't':
+            gap = strtol(optarg, NULL, 0);
+            if (gap < 1 || gap > GAP_MAX)
+                error("value for gap must be between 1 to %d", GAP_MAX);
             break;
         case 'r':
             linerate = strtol(optarg, NULL, 0);
             if (linerate < LINERATE_MIN || linerate > LINERATE_MAX)
                 error("value for rate must be between %d to %d", LINERATE_MIN, LINERATE_MAX);
-            bps = (uint64_t)linerate * 1000000;
-            tb.last_add = time(NULL);
+            if (linerate > 0)
+                bps = (uint64_t)linerate * 1000000;
             break;
         case 'h':
         default:
@@ -150,55 +154,36 @@ int main(int argc, char **argv)
         }
     }
 
+    /* don't use captured interval if any of the custom interval options is set */
+    if (pps > 0 || gap > 0 || linerate >= 0)
+        default_gap = false;
+
     if (device == NULL)
         error("device not specified");
 
     if (argv[optind] == NULL)
         error("trace file not specified");
 
-    notice("sending packets through %s", device);
-
-    /* empty error buffer to grab warning message (if exist) from pcap_open_live() below */
-    *ebuf = '\0';
-
-    /* note that we are doing this for sending packets, not capture */
-    pd = pcap_open_live(device, ETHER_MAX_LEN, /* portion of packet to capture */
-                        1,                     /* promiscuous mode is on */
-                        1000,                  /* read timeout, in milliseconds */
-                        ebuf);
-
-    if (pd == NULL)
-        error("%s", ebuf);
-    else if (*ebuf)
-        notice("%s", ebuf); /* warning message from pcap_open_live() above */
-
-    /* buffer to store data for each packet including its link-layer header, freed in cleanup() */
-    pkt_data = (uint8_t *)malloc(sizeof(uint8_t) * ETHER_MAX_LEN);
-    if (pkt_data == NULL)
-        error("malloc(): cannot allocate memory for pkt_data");
-    memset(pkt_data, 0, ETHER_MAX_LEN);
-
     /* set signal handler for SIGINT (Control-C) */
     (void)signal(SIGINT, cleanup);
 
+    load_trace_files(argc, argv);
+
+    init_pcap(device);
+
+    notice("sending packets through %s", device);
+
     if (gettimeofday(&start, NULL) == -1)
         notice("gettimeofday(): %s", strerror(errno));
 
-    trace_files = malloc((argc - optind) * sizeof(FILE *));
-    for (i = optind; i < argc; i++)
-    {
-        FILE *fp;
-        notice("trace file: %s", argv[i]);
-        if ((fp = fopen(argv[i], "rb")) == NULL)
-            error("fopen(): error reading %s", argv[i]);
-        trace_files[trace_files_count++] = fp;
-    }
+    clock_gettime(CLOCK_MONOTONIC, &tb_pps.last_add);
+    clock_gettime(CLOCK_MONOTONIC, &tb_linerate.last_add);
 
     /* send infinitely if loop <= 0 until user Control-C */
     while (1)
     {
         for (i = 0; i < trace_files_count; i++) /* for each trace file */
-            send_packets(device, trace_files[i], argv[i + optind]);
+            send_packets(trace_files[i]);
 
         if (loop > 1)
             loop--;
@@ -212,49 +197,144 @@ int main(int argc, char **argv)
     exit(EXIT_SUCCESS);
 }
 
-void send_packets(char *device, FILE *fp, char *trace_file)
+void load_trace_files(int argc, char **argv)
 {
-    struct pcap_file_header preamble;
-    struct pcap_sf_pkthdr header;
-    int pkt_len; /* packet length to send */
-    int ret;
-    int nsec = 0; /* set to 1 if we have timestamps in nanosecond resolution */
-    struct pcap_timeval p_ts;
-    struct timeval tv;
-    sigset_t block_sig;
+    struct pcap_file_header preamble; /* pcap file header per trace file */
 
-    (void)sigemptyset(&block_sig);
-    (void)sigaddset(&block_sig, SIGINT);
+    trace_files = malloc((argc - optind) * sizeof(trace_file_t *));
+    if (trace_files == NULL)
+        error("malloc(): cannot allocate memory for trace_files");
 
-    /* reset trace file pointer */
-    if (fseek(fp, 0, SEEK_SET) != 0)
-        error("fseek(): error reading %s", trace_file);
+    for (int i = optind; i < argc; i++)
+    {
+        trace_file_t *trace_file = malloc(sizeof(trace_file_t));
+        if (trace_file == NULL)
+            error("malloc(): cannot allocate memory for trace_file");
 
-    /* preamble occupies the first 24 bytes of a trace file */
-    if (fread(&preamble, sizeof(preamble), 1, fp) == 0)
-        error("fread(): error reading %s", trace_file);
-    if (preamble.magic != PCAP_MAGIC && preamble.magic != NSEC_PCAP_MAGIC)
-        error("%s is not a valid pcap based trace file", trace_file);
-    if (preamble.magic == NSEC_PCAP_MAGIC)
-        nsec = 1;
+        trace_file->filename = argv[i];
+
+        if ((trace_file->fp = fopen(trace_file->filename, "rb")) == NULL)
+            error("fopen(): error reading %s", trace_file->filename);
+
+        /*
+         * check preamble of each trace file.
+         * preamble occupies the first 24 bytes of a trace file
+         */
+        if (fread(&preamble, PCAP_PREAMBLE_LEN, 1, trace_file->fp) == 0)
+            error("fread(): error reading %s", trace_file->filename);
+
+        if (preamble.magic != PCAP_MAGIC && preamble.magic != NSEC_PCAP_MAGIC)
+            error("%s is not a valid pcap based trace file", trace_file->filename);
+
+        if (preamble.magic == NSEC_PCAP_MAGIC)
+            trace_file->nsec = true;
+        else
+            trace_file->nsec = false;
+
+        /* load all inter-packet gaps if we are using captured interval */
+        if (default_gap)
+            load_ipg(trace_file);
+        else
+            trace_file->ipg = NULL;
+
+        trace_files[trace_files_count++] = trace_file;
+    }
+}
+
+void load_ipg(trace_file_t *trace_file)
+{
+    struct timespec curr_ts;  /* timestamp of current packet */
+    struct timespec prev_ts;  /* timestamp of previous packet */
+    uint64_t pkts = 0;        /* packet index */
+    uint64_t i = 0;           /* ipg index */
+    uint64_t ipg_cap = 10000; /* initial capacity to store ipg values; doubled as needed */
+
+    trace_file->ipg = malloc(ipg_cap * sizeof(ipg_t));
+    if (trace_file->ipg == NULL)
+        error("malloc(): cannot allocate memory for ipg");
+
+    /*
+     * file pointer has moved past the pcap file header.
+     * loop through the remaining data by reading the packet header first.
+     * packet header (16 bytes) = timestamp + length
+     */
+    while (fread(&header, PCAP_HDR_LEN, 1, trace_file->fp) == 1)
+    {
+        curr_ts.tv_sec = header.ts.tv_sec;
+        if (trace_file->nsec)
+            curr_ts.tv_nsec = header.ts.tv_usec;
+        else
+            curr_ts.tv_nsec = header.ts.tv_usec * 1000;
+
+        /* move file pointer to the end of this packet data */
+        if (fseek(trace_file->fp, header.caplen, SEEK_CUR) != 0)
+            error("fseek(): error reading %s", trace_file->filename);
+
+        if (pkts > 0)
+        {
+            /* double the capacity to store ipg values if current capacity is hit */
+            if (i >= ipg_cap)
+            {
+                ipg_cap *= 2;
+                ipg_t *new_ipg = realloc(trace_file->ipg, ipg_cap * sizeof(ipg_t));
+                if (new_ipg == NULL)
+                    error("realloc(): cannot allocate memory for new_ipg");
+                trace_file->ipg = new_ipg;
+            }
+
+            trace_file->ipg[i].ns = (curr_ts.tv_sec - prev_ts.tv_sec) * 1000000000L +
+                                    (curr_ts.tv_nsec - prev_ts.tv_nsec);
+
+            /* pps value dictates how next packet is to be throttled during the send */
+            if (trace_file->ipg[i].ns == 0)
+                trace_file->ipg[i].pps = 0; /* send immediately */
+            else if (trace_file->ipg[i].ns <= 1000000000)
+                trace_file->ipg[i].pps = 1000000000 / trace_file->ipg[i].ns; /* throttle at pps */
+            else
+                trace_file->ipg[i].pps = -1; /* pps is less than 1, fallback to sleep for ns */
+
+            ++i;
+        }
+
+        ++pkts;
+
+        prev_ts = curr_ts;
+    }
+}
+
+void init_pcap(char *device)
+{
+    /* empty error buffer to grab warning message (if exist) from pcap_open_live() below */
+    *ebuf = '\0';
+
+    /* note that we are doing this for sending packets, not capture */
+    pd = pcap_open_live(device, ETH_MAX_LEN, /* portion of packet to capture */
+                        1,                   /* promiscuous mode is on */
+                        1000,                /* read timeout, in milliseconds */
+                        ebuf);
+
+    if (pd == NULL)
+        error("%s", ebuf);
+    else if (*ebuf)
+        notice("%s", ebuf); /* warning message from pcap_open_live() above */
+}
+
+void send_packets(trace_file_t *trace_file)
+{
+    int pkt_len;       /* packet length to send */
+    struct timeval tv; /* current timestamp for verbose output */
+    uint64_t pkts = 0; /* packet index */
+
+    /* reset trace file pointer moving past the pcap file header */
+    if (fseek(trace_file->fp, PCAP_PREAMBLE_LEN, SEEK_SET) != 0)
+        error("fseek(): error reading %s", trace_file->filename);
 
     /*
      * loop through the remaining data by reading the packet header first.
      * packet header (16 bytes) = timestamp + length
      */
-    while ((ret = fread(&header, sizeof(header), 1, fp)))
+    while (fread(&header, PCAP_HDR_LEN, 1, trace_file->fp) == 1)
     {
-        if (ret == 0)
-            error("fread(): error reading %s", trace_file);
-
-        /* copy timestamp for current packet */
-        memcpy(&p_ts, &header.ts, sizeof(p_ts));
-        curr_ts.tv_sec = p_ts.tv_sec;
-        if (nsec == 1)
-            curr_ts.tv_nsec = p_ts.tv_usec;
-        else
-            curr_ts.tv_nsec = p_ts.tv_usec * 1000;
-
         if (len < 0) /* captured length */
             pkt_len = header.caplen;
         else if (len == 0) /* actual length */
@@ -262,13 +342,12 @@ void send_packets(char *device, FILE *fp, char *trace_file)
         else /* user specified length */
             pkt_len = len;
 
-        throttle(pkt_len);
+        /* skip throttling if linerate is set to 0, i.e. send each packet immediately */
+        if (linerate != 0)
+            throttle(trace_file, pkts, pkt_len * 8);
 
-        load_packet(fp, pkt_len, &header, trace_file);
+        load_packet(trace_file, pkt_len, &header);
 
-        (void)sigprocmask(SIG_BLOCK, &block_sig, NULL); /* hold SIGINT */
-
-        /* finish the injection and verbose output before we give way to SIGINT */
         if (pcap_sendpacket(pd, pkt_data, pkt_len) == -1)
         {
             notice("%s", pcap_geterr(pd));
@@ -279,9 +358,6 @@ void send_packets(char *device, FILE *fp, char *trace_file)
             ++pkts_sent;
             bytes_sent += pkt_len;
 
-            /* copy timestamp for previous packet sent */
-            memcpy(&prev_ts, &curr_ts, sizeof(struct timespec));
-
             /* verbose output */
             if (vflag)
             {
@@ -301,65 +377,96 @@ void send_packets(char *device, FILE *fp, char *trace_file)
             }
         }
 
-        (void)sigprocmask(SIG_UNBLOCK, &block_sig, NULL); /* release SIGINT */
+        ++pkts;
 
         if ((max_pkts > 0) && (pkts_sent >= max_pkts))
             cleanup(0);
     } /* end while */
 }
 
-void throttle(int pkt_len)
+void sleep_ns(uint64_t ns)
 {
+    /* cumulative drift to adjust ns accordingly based on previous elapsed_ns */
+    static int64_t drift_ns = 0;
+
+    struct timespec ts, rem_ts, start_ts, end_ts;
+    ts.tv_sec = (ns + drift_ns) / 1000000000;
+    ts.tv_nsec = (ns + drift_ns) % 1000000000;
+
+    clock_gettime(CLOCK_MONOTONIC, &start_ts);
+    while (nanosleep(&ts, &rem_ts) == -1 && errno == EINTR)
+        ts = rem_ts;
+    clock_gettime(CLOCK_MONOTONIC, &end_ts);
+
+    uint64_t elapsed_ns =
+        (end_ts.tv_sec - start_ts.tv_sec) * 1000000000 + (end_ts.tv_nsec - start_ts.tv_nsec);
+
     /*
-     * always send first packet immediately.
-     * if linerate is set to 0, send each packet immediately
+     * if elapsed_ns > ns, the negative drift will shorten the next sleep_ns().
+     * if elapsed_ns < ns, the positive drift will lengthen the next sleep_ns()
      */
-    if (linerate == 0 || !timespecisset(&prev_ts))
+    drift_ns += ns - elapsed_ns;
+}
+
+void throttle(trace_file_t *trace_file, uint64_t pkts, int bits)
+{
+    /* always send first packet immediately */
+    if (pkts_sent == 0)
         return;
 
-    if (interval > 0) /* user specified interval is in seconds only */
-        sleep_ts.tv_sec = interval;
+    if (pps)
+    {
+        /* throttle using token bucket algorithm if pps is specified */
+        while (!token_bucket_remove(&tb_pps, 1, pps))
+            usleep(1);
+    }
+    else if (bps)
+    {
+        /* throttle using token bucket algorithm if linerate is specified */
+        while (!token_bucket_remove(&tb_linerate, bits, bps))
+            usleep(1);
+    }
+    else if (gap) /* user specified inter-packet gap in seconds */
+    {
+        sleep_ns(gap * 1000000000L);
+    }
     else /* use captured interval */
     {
         /*
-         * fallback to last sleep_ts if curr_ts < prev_ts which can happen when looping over
-         * multiple trace file
+         * note that the first packet in this trace file is always sent immediately.
+         * this applies even if we are looping the same trace file multiple times
          */
-        if (timespeccmp(&curr_ts, &prev_ts, >=))
-            timespecsub(&curr_ts, &prev_ts, &sleep_ts);
-    }
-
-    if (bps)
-    {
-        /* throttle using token bucket algorithm if linerate is specified */
-        if (token_bucket_remove(&tb, pkt_len * 8, bps))
-            return;
-    }
-    else if (timespecisset(&sleep_ts))
-    {
-        if (sleep_ts.tv_sec > SLEEP_MAX)
-            notice("warning: next packet has timestamp over %lu seconds away", sleep_ts.tv_sec);
-        if (nanosleep(&sleep_ts, NULL) == -1)
-            notice("nanosleep(): %s", strerror(errno));
+        if (pkts > 0)
+        {
+            if (trace_file->ipg[pkts - 1].pps == 0) /* send immediately */
+                return;
+            else if (trace_file->ipg[pkts - 1].pps < 0) /* sleep for ns */
+                sleep_ns(trace_file->ipg[pkts - 1].ns);
+            else /* throttle using token bucket algorithm at pps */
+            {
+                while (!token_bucket_remove(&tb_pps, 1, trace_file->ipg[pkts - 1].pps))
+                    usleep(1);
+            }
+        }
     }
 }
 
-void load_packet(FILE *fp, int pkt_len, struct pcap_sf_pkthdr *header, char *trace_file)
+void load_packet(trace_file_t *trace_file, int pkt_len, struct pcap_sf_pkthdr *header)
 {
     int copy_len = pkt_len < header->caplen ? pkt_len : header->caplen;
 
-    if (fread(pkt_data, sizeof(char), copy_len, fp) != copy_len)
-        error("fread(): error reading %s", trace_file);
+    if (fread(pkt_data, 1, copy_len, trace_file->fp) != copy_len)
+        error("fread(): error reading %s", trace_file->filename);
 
     /* pad trailing bytes with zeros */
-    for (int i = copy_len; i < pkt_len; i++)
-        pkt_data[i] = PKT_PAD;
+    if (copy_len < pkt_len)
+        memset(pkt_data + copy_len, PKT_PAD, pkt_len - copy_len);
 
     /* move file pointer to the end of this packet data */
     if (copy_len < header->caplen)
     {
-        if (fseek(fp, header->caplen - copy_len, SEEK_CUR) != 0)
-            error("fseek(): error reading %s", trace_file);
+        if (fseek(trace_file->fp, header->caplen - copy_len, SEEK_CUR) != 0)
+            error("fseek(): error reading %s", trace_file->filename);
     }
 }
 
@@ -367,7 +474,7 @@ void info(void)
 {
     struct timeval elapsed;
     float seconds;
-    unsigned long bits_sent;
+    unsigned long bits_sent, actual_pps;
     float mbps, gbps;
 
     if (gettimeofday(&end, NULL) == -1)
@@ -375,13 +482,14 @@ void info(void)
     timersub(&end, &start, &elapsed);
     seconds = elapsed.tv_sec + (float)elapsed.tv_usec / 1000000;
 
+    actual_pps = pkts_sent / seconds;
     bits_sent = bytes_sent * 8;
     mbps = bits_sent / seconds / 1000000;
     gbps = bits_sent / seconds / 1000000000;
 
     (void)putchar('\n');
-    notice("%lu packets (%lu bytes) sent", pkts_sent, bytes_sent);
-    notice("throughput = %.4f Mbps (%.4f Gbps)", mbps, gbps);
+    notice("sent = %lu packets, %lu bits, %lu bytes", pkts_sent, bits_sent, bytes_sent);
+    notice("throughput = %lu pps, %.4f Mbps, %.4f Gbps", actual_pps, mbps, gbps);
     if (failed)
         notice("%lu write attempts failed", failed);
     notice("elapsed time = %f seconds", seconds);
@@ -389,11 +497,12 @@ void info(void)
 
 void cleanup(int signum)
 {
-    free(pkt_data);
-    pkt_data = NULL;
-
     for (int i = 0; i < trace_files_count; i++)
-        fclose(trace_files[i]);
+    {
+        fclose(trace_files[i]->fp);
+        if (trace_files[i]->ipg != NULL)
+            free(trace_files[i]->ipg);
+    }
     free(trace_files);
     trace_files = NULL;
 
@@ -528,36 +637,38 @@ void error(const char *fmt, ...)
 void usage(void)
 {
     (void)fprintf(stderr,
-                  "%s version %s\n"
+                  "%s version %s, Copyright (C) 2006 - 2023 Addy Yeow <ayeowch@gmail.com>\n"
                   "%s\n"
                   "Usage: %s [-d] [-v] [-i interface] [-s length] [-l loop] [-c count]\n"
-                  "                [-p sleep] [-r rate] [-h] pcap-file(s)\n"
+                  "                [-p pps] [-t gap] [-r rate] [-h] pcap-file(s)\n"
                   "\nOptions:\n"
                   " -d             Print a list of network interfaces available.\n"
                   " -v             Print timestamp for each packet.\n"
                   " -vv            Print timestamp and hex data for each packet.\n"
                   " -i interface   Send 'pcap-file(s)' out onto the network through 'interface'.\n"
                   " -s length      Packet length to send (in bytes). Set 'length' to:\n"
-                  "                     0 to send the actual packet length. This is the default.\n"
-                  "                    -1 to send the captured length.\n"
+                  "                 0 : Send the actual packet length. This is the default.\n"
+                  "                -1 : Send the captured length.\n"
                   "                or any other value from %d to %d.\n"
                   " -l loop        Send 'pcap-file(s)' out onto the network for 'loop' times.\n"
                   "                Set 'loop' to 0 to send 'pcap-file(s)' until stopped.\n"
                   "                To stop, type Control-C.\n"
                   " -c count       Send up to 'count' packets.\n"
                   "                Default is to send all packets from 'pcap-file(s)'.\n"
-                  " -p sleep       Set interval to 'sleep' (in seconds), ignoring the actual\n"
-                  "                interval.\n"
-                  "                Value for 'sleep' must be between 1 to %d.\n"
+                  " -p pps         Send 'pps' packets per second.\n"
+                  "                Value for 'pps' must be between 1 to %d.\n"
+                  " -t gap         Set inter-packet 'gap' in seconds, ignoring the captured\n"
+                  "                interval between packets in 'pcap-file(s)'.\n"
+                  "                Value for 'gap' must be between 1 to %d.\n"
                   " -r rate        Limit the sending to 'rate' Mbps.\n"
                   "                Value for 'rate' must be between %d to %d.\n"
-                  "                This option is intended to set the desired packet throughput.\n"
+                  "                This flag is intended to set the desired packet throughput.\n"
                   "                If you want to send packets at line rate of 1000 Mbps,\n"
                   "                try -r 1000. If you want to send packets without rate limit,\n"
-                  "                try -r 0. This option is typically used with -l 0 to send\n"
+                  "                try -r 0. This flag is typically used with -l 0 to send\n"
                   "                'pcap-file(s)' until stopped.\n"
                   " -h             Print version information and usage.\n",
-                  program_name, BITTWIST_VERSION, pcap_lib_version(), program_name, ETHER_HDR_LEN,
-                  ETHER_MAX_LEN, SLEEP_MAX, LINERATE_MIN, LINERATE_MAX);
+                  program_name, BITTWIST_VERSION, pcap_lib_version(), program_name, ETH_HDR_LEN,
+                  ETH_MAX_LEN, PPS_MAX, GAP_MAX, LINERATE_MIN, LINERATE_MAX);
     exit(EXIT_SUCCESS);
 }
