@@ -2,7 +2,7 @@
  * SPDX-License-Identifier: GPL-2.0-or-later
  *
  * token_bucket - token bucket algorithm
- * Copyright (C) 2006 - 2023 Addy Yeow Chin Heng <ayeowch@gmail.com>
+ * Copyright (C) 2006 - 2023 Addy Yeow <ayeowch@gmail.com>
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -23,22 +23,25 @@
 
 void token_bucket_add(struct token_bucket *tb, uint64_t bps)
 {
-    time_t now = time(NULL);
-    tb->tokens = tb->tokens + (bps * (now - tb->last_add));
-    tb->tokens = (tb->tokens > bps) ? bps : tb->tokens;
+    struct timespec now;
+    clock_gettime(CLOCK_MONOTONIC, &now);
+
+    uint64_t elapsed_ns =
+        (now.tv_sec - tb->last_add.tv_sec) * 1000000000 + (now.tv_nsec - tb->last_add.tv_nsec);
+
+    tb->tokens += bps * (double)elapsed_ns / 1000000000;
+    tb->tokens = (tb->tokens < bps) ? tb->tokens : bps;
     tb->last_add = now;
 }
 
 bool token_bucket_remove(struct token_bucket *tb, uint64_t bits, uint64_t bps)
 {
     if (tb->tokens < bits)
-    {
-        usleep(1);
         token_bucket_add(tb, bps);
-    }
 
     if (tb->tokens < bits)
         return false;
+
     tb->tokens -= bits;
     return true;
 }
